#  File - rmsb02.ch
#  This file describes the changes from RMS B.01.02 to RMS B.02.00 tables 
#  The tblmnt utility is run against this file, reads the changes,
#  and acts on those.
#
#  The '#' character indicates a comment line and MUST begin at the
#  first position in the line so as not to be confused with a column
#  name or table name.  Ingres allows this character in the naming
#  rules.
#
#  See /usr/ingres/tblmnt for the source code
#  or try tblmnt -help for a list of options

TABLE_NAME	calc_start_dates
	drop_table
end_table_changes

TABLE_NAME	fixed_call_matrix
	drop_table
end_table_changes

TABLE_NAME	fixed_tat_matrix
	drop_table
end_table_changes

TABLE_NAME	ign_call_status_matrix
	drop_table
end_table_changes

TABLE_NAME	priority_matrix
	drop_table
end_table_changes

TABLE_NAME	wc_call_matrix
	drop_table
end_table_changes

TABLE_NAME	wc_tat_matrix
	drop_table
end_table_changes

# Map marketing services to production services
# These will be comma separated lists
TABLE_NAME	map_mkt_to_prod
        drop_table
	create_table
	addcol	mkt_srv_combos varchar(50) notnull notdefault
	addcol  sequence integer notnull notdefault
	addcol  repair_class varchar(2) notnull notdefault
	addcol  srv_type varchar(03) notnull notdefault
	addcol  srv_sub_type varchar(20) notnull notdefault
	addcol  first_phase integer notnull notdefault
	addcol  tat_code varchar(02) withnull
	post_sql copy map_mkt_to_prod (\
		 mkt_srv_combos = varchar(0)tab,\
		 sequence = c0tab,\
		 repair_class = varchar(0)tab,\
		 srv_type = varchar(0)tab,\
		 srv_sub_type = varchar(0)tab,\
		 first_phase = c0tab,\
		 tat_code = varchar(0)nl,\
		 nl=d1) from 'map_mkt.ing'
end_table_changes

TABLE_NAME	csc_defaults
	delcol	calc_commit
	delcol 	commitment_days
	delcol 	default_priority
	delcol 	override_commit
	delcol 	rms_routing
	delcol 	tech_ab_proc
	delcol 	use_calltype
	delcol 	use_priority
	delcol 	use_reptype
	rename  start_days ship_buffer_days
end_table_changes

TABLE_NAME	emp
	delcol emp_location
	delcol emp_phone_no
	delcol emp_skills1
	delcol emp_skills2
	delcol entity_code
	delcol loc_code
end_table_changes

# Do this so that product data will be copied out in case of problems
# Only want to copy out the first time it is run
TABLE_NAME	product
	rename cal_smp cal_sup
	rename repair_smp repair_sup
	post_sql drop table product_service
	post_sql create table product_service as select * from product
end_table_changes

# Errors ok on insert statement if this is run twice
TABLE_NAME	product
        drop_table
	create_table
	addcol model_no varchar(20) notnull notdefault
	addcol srv_type varchar(03) notnull notdefault
	addcol labor_hours float notnull notdefault
	addcol jobs integer notnull notdefault
	addcol avg_hours float notnull notdefault
	addcol grp_hours float notnull notdefault
	addcol wc_id varchar(08) withnull 
	addcol system varchar(20) withnull
	addcol sup float withnull
	post_sql insert into product select model_no, 'CAL', labor_hours_cal,\
		 jobs_cal, avg_hours_cal, grp_hours_cal, cal_resource, \
		 cal_system, cal_sup from product_service where jobs_cal > 0
	post_sql insert into product select model_no, 'REP', labor_hours_rep,\
		 jobs_rep, avg_hours_rep, grp_hours_rep, rep_resource, \
		 NULL, repair_sup from product_service where jobs_rep > 0
	post_sql drop table product_service
	post_sql modify product to btree unique on model_no, srv_type
end_table_changes

TABLE_NAME	sups
	resize srv_type 3
	post_sql	update sups set srv_type='CAL' where srv_type='C'
	post_sql	update sups set srv_type='REP' where srv_type='R'
end_table_changes

TABLE_NAME	valid_repair_classes
	resize srv_type 3
	post_sql	update valid_repair_classes \
			set srv_type='CAL' where srv_type='C'
	post_sql	update valid_repair_classes \
			set srv_type='REP' where srv_type='R'
end_table_changes

TABLE_NAME	valid_services
        drop_table
	create_table 
	addcol	srv_type varchar(03) notnull notdefault
	addcol	prod_cap varchar(01) notnull notdefault
	addcol	srv_descr varchar(40) withnull
	post_sql insert into valid_services values ('REP','Y','Repair')
	post_sql insert into valid_services values ('CAL','Y','Calibration')
	post_sql insert into valid_services values ('OTH','N','Other')
end_table_changes

# Do this so that data will be copied out in case of problems
# Only want to copy out the first time it is run
TABLE_NAME	 work_center
	rename thru_put thru_delete
	post_sql drop table work_center_service
	post_sql create table work_center_service as select * from work_center
end_table_changes

TABLE_NAME	work_center
        drop_table
	create_table 
	addcol wc_id varchar(08) notnull notdefault
	addcol srv_type varchar(03) notnull notdefault
	addcol start_days integer notnull notdefault
	addcol mpj integer notnull notdefault
	post_sql insert into work_center \
		 select wc_id, 'CAL', wc_start_days, cal_mpj \
		 from work_center_service where cal_mpj > 0
	post_sql insert into work_center \
		 select wc_id, 'REP', wc_repair_days, repair_mpj \
		 from work_center_service where repair_mpj > 0
	post_sql insert into work_center \
		 select wc_id, 'OTH', wc_repair_days, repair_mpj \
		 from work_center_service where srv_type='O' or srv_type='OTH'
	post_sql drop table work_center_service
	post_sql modify work_center to btree unique on wc_id, srv_type
end_table_changes

TABLE_NAME	work_history
	resize srv_type 3
	delcol commit_warning
	delcol srv_priority
	delcol wc_ovr
	delcol labor_ovr
	delcol misc_charge1
	delcol misc_charge2
	delcol misc_charge3
	delcol misc_charge4
	delcol msg_type
	addcol caller_name varchar(24) withnull
	addcol caller_phone varchar(24) withnull
	addcol base_cso varchar(12) withnull
	rename commit_date sched_date
	rename resource cal_system
	post_sql	set nologging 
	post_sql	update work_history \
			set srv_type='CAL' where srv_type='C' 
	post_sql	update work_history \
			set srv_type='REP' where srv_type='R' 
end_table_changes

#	post_sql	update work_history \
#			set srv_type='REP' where srv_type='R' \
#			and tid >= (select max(tid)/2 from work_history)
TABLE_NAME	work_product
	resize srv_type 3
	delcol commit_warning
	delcol srv_priority
	delcol wc_ovr
	delcol labor_ovr
	delcol misc_charge1
	delcol misc_charge2
	delcol misc_charge3
	delcol misc_charge4
	delcol msg_type
	addcol caller_name varchar(24) withnull
	addcol caller_phone varchar(25) withnull
	addcol base_cso varchar(12) withnull
	rename commit_date sched_date
	rename resource cal_system
	post_sql	update work_product \
			set srv_type='CAL' where srv_type='C'
	post_sql	update work_product \
			set srv_type='REP' where srv_type='R'
end_table_changes

TABLE_NAME	work_orphans
	resize srv_type 3
	delcol commit_warning
	delcol srv_priority
	delcol wc_ovr
	delcol labor_ovr
	delcol misc_charge1
	delcol misc_charge2
	delcol misc_charge3
	delcol misc_charge4
	delcol msg_type
	addcol caller_name varchar(24) withnull
	addcol caller_phone varchar(24) withnull
	addcol base_cso varchar(12) withnull
	rename commit_date sched_date
	rename resource cal_system
	post_sql	update work_orphans \
			set srv_type='CAL' where srv_type='C'
	post_sql	update work_orphans \
			set srv_type='REP' where srv_type='R'
end_table_changes

# This table is dependent on work_product changes above
TABLE_NAME	fmn_load
	drop_table
	post_sql	create table fmn_load as select * from work_product \
			where tid in (select min(tid) from work_product)
end_table_changes

@EOF
set `wc -lwc <02rdf57`
if test $1$2$3 != 2488447385
then
	echo ERROR: wc results of 02rdf57 are $* should be 248 844 7385
fi

chmod 660 02rdf57

exit 0

