/*
| TBLMNT - add, delete, rename, resize columns in an Ingres table.
| ----------------------------------------------------------------
|           Copyright (C) 1995   Author: Ken Frank
|               v1.00 released on 13-apr-1995
|                       This is v1.21
| ----------------------------------------------------------------
| This program is free software; you can redistribute it and/or
| modify it under the terms of the GNU General Public License
| as published by the Free Software Foundation; either version 2
| of the license, or (at your option) any later version.
|
| This program is distributed in the hope that it will be useful,
| but WITHOUT ANY WARRANTY; without even the implied warranty of
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
| GNU General Public License for more details.
|
| You should have received a copy of the GNU General Public License
| along with this program; if not, write to the Free Software
| Foundation, Inc., 675 Mass Ave, Cambridge, MA, 02139, USA.
*/

/*
| Compiling instructions:
| For all cases, run the pre-compiler first:
|       esqlc tblmnt.sc
|
| Then, depending on your hardware platform, pick the appropriate 'C'
| compiler directions:
| -------------------------------------------------------------------
| NT 3.51 (MS C++ v4.0):
|  cl tblmnt.c /D NO_CUSERID -link %II_SYSTEM%\ingres\lib\esqlc.lib
|
|
| HP-UX:
|  cc -Aa tblmnt.c -L$II_SYSTEM/ingres/lib -lingres -lc -lm -o tblmnt
|
|
| AIX:
|  cc tblmnt.c -L$II_SYSTEM/ingres/lib -lingres -lc -lm -o tblmnt
|
|
| UTS (Amdahl Unix):
|  cc tblmnt.c -g -Xc -eft -L$II_SYSTEM/ingres/lib -lingres -lsocket -o tblmnt
|
|
| VMS:
|  cc tblmnt.c
|  link tblmnt.obj, II_SYSTEM:[ingres.files]esql.opt/opt /exe=tblmnt.exe
|
|
| OpenVMS:
| cc tblmnt.c /standard=common
| link tblmnt.obj, ii_system:[ingres.files]esql.opt/opt  /exe=tblmnt.exe
|
|
| SCO:
| cc tblmnt.c -L$II_SYSTEM/ingres/lib -lingres1 -lingres2 -lsocket -lc -lm \
|                                               -o tblmnt
*/


/*
| program:
|       tblmnt.sc
|
| purpose:
|       This program is indended to provide a safe and reliable method
|       for maintaining the columns in a table.
|
| commands supported:
|       addcol          - specifies column and definition to be added to table
|       delcol          - specifies column to be dropped from table
|       resize          - specifies new length for column
|       resize_numeric  - specifies new numeric storage size
|       rename          - rename an existing column
|       withnull        - make a column nullable
|       withdefault     - make a column defaulted
|       notnull         - make a column not nullable
|       notdefault      - make a column not defaulted
|       sort_col_names  - sort the column-names (so HELP table is sorted)
|       colseq          - explicitly specifies a column sequence.
|       create_table    - indicates that this is a NEW table (don't try to save)
|       drop_table      - drop an existing table, no abort if it doesn't exist
|       keep_prikey     - the primary-key/table structure will be kept
|       keep_altkeys    - the secondary key structures will be kept
|       keep_rules      - any rules for the table will be kept
|       keep_views      - any views for the table will be kept
|       keep_integrities- any integrities for the table will be kept
|       keep_permits    - any permits for the table will be kept
|       keep_journalling- option for preserving journaling status of table
|       post_sql        - any Ingres SQL statement (except for 'select')
*/

/*
| Modification History:
| ---------------------
| 13-apr-1995   1.00    Ken Frank       -- initial release
| 17-apr-1995   1.01    Ken Frank       -- fixed a couple of minor printf bugs.
| 25-apr-1995   1.02    Ken Frank
|               I messed up in the section that restores secondary indexes.
|               The build_alt_index_stmt() routine was referencing TBL_ARR[0],
|               instead of TBL_ARR[ix_no] when checking to see if a "leaffill"
|               should be put on the index structure.  Also, in this same
|               routine, I was specifying the .lfillpct value instead of the
|               .dfillpct for the "fillfactor" parameter.
|
|               Also, if a HASH structure had a value of zero in
|               iitables.table_minpages, the key was not re-created because
|               Ingres complains that "zero is an illegal value for minpages"
|               [so how'd it get that way in the first place?]
|
| 26-apr-1995   1.03    Ken Frank
|               The "notnull" command was broken - it was correctly specifying
strcpy(the_table, tmp);ll('0') [or withnull('')] on the copyout, but on the
|               reload, should not have had ANY withnull() option.
|
| 05-may-1995   1.04    Ken Frank
|               Added support for "-uuser", though it must still be the 4th
|               parameter.  Handle "blank" commands as if they were simply
|               empty lines.
|
| 08-may-1995   1.05    Ken Frank
|               Apparently, columns with a data-type of TABLE_KEY aren't
|               accurately reported in IIcolumns.column_datatype, but instead
|               must be checked in IIcolumns.column_internal_datatype.  Also,
|               when rules were saved, an initial blank line was sometimes
|               written out, causing an error-message - even though all the
|               rules did get properly restored.
|
| 10-may-1995   1.06    Ken Frank
|               I was being careless with the strtok() function, and assuming
|               that when it returned NULL, I could test this by checking if
|               the result was '\0'.  It worked fine on UNIX, but VMS didn't
|               like it, and gave me stack dumps.  This was only a problem
|               when lines with nothing but white-space were present in the
|               script-file, or on ill-formed commands.
|
| 16-may-1995   1.07    Ken Frank
|               I finally learned what "ret_to_all", in iitables means.  The
|               "grant select on <table> to public" is now preserved.
|               (Thanks Mike Ashline - mashline@m1mel.mis.semi.harris.com)
|               Also, somehow (?), the bug I fixed in v1.02 (checking table-
|               structure for base table, rather than current index) crept
|               back into the source code.  I have no idea how that happened,
|               but everything else appears to be ok.
|
| 28-jul-1995   1.08    Ken Frank
|               Added a commit after the table-grants were restored (and
|               after the table was reloaded).  Previously, if the first
|               post_sql statement failed, it would roll back the grants,
|               leaving the table with no permissions.  Added support for
|               the following command-line option:
|                 "-fcleanup"  --> Clean up all the intermediate files created
|                                  except the .log file.  This will override
|                                  any "keep_copyoutfile = y" command in the
|                                  script.  Any intermediate files which could
|                                  not be processed cleanly, however, will not
|                                  be deleted.
|
| 22-aug-1995   1.09    Ken Frank
|               After hearing a number of times that it is faster to modify
|               the table while it is empty and reload it (rather than to
|               reload it, and then modify it), that is now the default
|               behaviour.  You can still get the "old" behaviour with a
|               "modify_before_reload = N" statement in your script.
|               For what it's worth, the time-difference doesn't appear to
|               be very significant (though I've only played with some
|               rather small [ie < 1000 pages] tables).
|
| 06-oct-1995   1.10    Ken Frank
|               Finally made the parser a little smarter in handling
|               continued lines.  Now, a line is considered to be
|               continued if the last non-whitespace character on
|               a line is "\".  Also, (and more importantly), secondary
|               indexes are now re-created with "data columns" and
|               compression.
|
| 10-oct-1995   1.11    Ken Frank
|               Aaaiiighh!  IBM's man-page for the memset function
|               tells you to include memory.h, which is obsolete, and
|               non-existant on our VAX.  It is supposed to be string.h
|
| 07-nov-1995   1.12    Ken Frank
|               Views owned by people OTHER THAN the table-owner are now
|               re-created, under the proper ownership and everything!
|
| 08-nov-1995   1.13    Ken Frank
|               Missed a new parameter to save_permits() routine.  Worked
|               fine on UNIX, but once again, the old VAX choked on it.
|
| 10-nov-1995   1.14    Ken Frank
|               Made some "defaults" different (handling of duplicate or
|               non-existant columns), as well as the panic-messages.
|               Also, jmorr@hookup.net (Joanne Morrison) found a problem
|               (and the fix as well - thanks!):  the checkerrors()
|               routine was flushing the logf file before the logf had
|               been fopen'ed.  Now, the file is opened before any DB
|               interaction is performed.  (Amazingly, it worked on VAX).
|
| 13-feb-1996   1.15    Ken Frank and Per Oyvind Overli
|               Lots of neat stuff in this release!
|               Per contributed some code to provide the following
|               features (which I merged and mangled):
|                   1)  "colseq" command (specify ordering of a column
|                          in the table's definition.
|                   2)  all views (even if only indirectly dependant
|                          on the base table) are now saved/restored.
|                          [recursion w/ SQL is harder than you'd think]
|
|               And some things I did:
|                   3) Added (finally) the "sort_col_names = Y" command.
|                       (caused me to re-do most of Per's colseq stuff).
|                   4) Changed the "modify_before_reload" default to "N".
|                       Ingres 6.4 does *way* too much logging when loading
|                       an ISAM table from a file.  If you're a die-hard
|                       believer in modifying your table BEFORE loading it,
|                       just add a line "modify_before_reload = Y" into
|                       your tblmnt command-script.
|
|
| 14-mar-1996   1.16    Ken Frank
|               Make sure that all the views are recreated in the required
|               order, by using the view's reltid for sorting purposes.
|               Also, looking ahead to OpenIngres, I've added support for
|               the new datatypes (DECIMAL and BYTE [no security_keys]),
|               and column default-values.  I'd really appreciate getting
|               feedback from anyone using this on OpenIngres!
|
| 06-sep-1996   1.17    Ken Frank, and
|                       Greg Blackwell (gblackwe@vprpmel1.telstra.com.au)
|               Added new flag to make restoring JOURNALING optional
|               (because journalled tables use the Ingres transaction
|                log when you load them from a file)
|
|               The new flag works as follows:
|
|               keep_journalling = Y    (default)
|               keep_journalling = N    (table will no longer be journalled)
|               keep_journalling = R    (journalling re-enabled after next
|                                        checkpoint).
|
|               If you choose the 'R' option, you should be aware that
|               the table won't really be journalled until you take your
|               next checkpoint.
|               Thanks to Greg Blackwell (gblackwe@vprpmel1.telstra.com.au)
|               for input.
|
|
| 06-sep-1996   1.18    Bill Collins, Bob Miller, and Ken Frank
|               1) Bill (collins@telcores.com) found a couple of
|                  minor problems: OpenIngres column-defaults
|                  weren't preserved, and compressed-HEAP tables
|                  lost their compression.  Thanks for the code,
|                  Bill!  (And also provided some much-needed
|                  commiseration with OpenRoad).
|               2) Bob (miller@telcores.com) was forced by his
|                  picky SUN 'C' compiler to ANSI-ify the tblmnt
|                  source code, and let me use it.
|               3) I added a minor kludge to work around the fact
|                  that Windows doesn't have the `cuserid()' function
|                  available.  (The downside is that when running
|                  tblmnt under Windows, you are now required to
|                  use the -uUSER option).  Actually, the stupid
|                  compiler-trick I was using doesn't work as I'd
|                  hoped (#ifdef cuserid...), so for now, you'll
|                  need to use /D NO_CUSERID when compiling for
|                  Windows platforms.  If anyone knows a better
|                  way to get around this, please tell me!
|
| 06-nov-1996   1.19   Dennis Roesler  (Dennis_Roesler@hp.com)
|               Added the resize_numeric keyword.  This supports
|               converting a lesser numeric storage size to a larger
|               numeric storage size.
|
| 08-nov-1996   1.20    Dennis Roesler (Dennis_Roesler@hp.com)
|               Added the table_name/end_table_changes keywords.
|               The functionality of this is to allow multiple table
|               changes from one file, instead of one file per table
|               change.  These keywords are required keywords.  Each
|               set of table changes is acted on as an autonomous
|               transaction.
|
|               Also, allows for comments to appear at the beginning
|               of the file.
|
|               Set curr_user at the same place in the code as where
|               the_owner is set and commented out where curr_user
|               was being set to '\0' in what is now a while loop.
|               This was causing some problems doing post_sql commands.
|               You can have multiple post_sql commands within the
|               same set of table_name/end_table_changes keywords.
|
|               Added a -h or -help switch to get a list of directives
|               and there usage.  I.e., tblmnt -h or tblmnt -help.
|
| 21-mar-1997   1.21    Leslie Shkolnik
|               Added drop_table command so that you can drop a table before
|               creating it and it doesn't matter if it doesn't already
|               exist.
|
|               Added null specification and default specification on same
|               line as addcol.
|
|               Changed to not panic if not a new table and it doesn't exist
|               just warn and continue
|
|               Initialize _saved vars, view_args for each table block.
|               Was trying to restore views from a previous table.
|
| $Date: 97/03/25 16:27:29 $
| $Header: tblmnt.sc,v 1.4 97/03/25 16:27:29 ingres Exp $
| $Revision: 1.4 $
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

EXEC SQL include sqlca;

#define PROG_NAME               "tblmnt v1.21"
#define MAX_COLS_PER_TABLE      301
#define MAX_INDEXES             40
#define MAX_VIEWS               100
#define NONKEY_seq              -10
#define AND                     &&
#define OR                      ||
#define TRUE                    1
#define FALSE                   0


EXEC SQL begin declare section;
int     ingrows, ingerr;
char    ingmsg[512], msg[512];
char    the_datafilename[256];
char    the_logfilename[256];
char    the_rulefilename[256];
char    the_viewfilename[256];
char    the_integfilename[256];
char    the_permfilename[256];
char    the_psqlfilename[256];
char    the_db[60], the_table[60], the_script[256], the_owner[256];
char    curr_user[40];  /* who are we currently connected to database as */
char    CleanUp = 'N';
char    STATEMENT[40000];
int     rules_saved = 0;
int     integrities_saved = 0;
int     views_saved = 0;
int     permits_saved = 0;
int     post_sql_saved = 0;

/*
| The COL_REC typedef defines a structure containing all the columns in a table.
*/
typedef struct {
        char    name[25], new_name[25];
        char    type[33];
        int     size, new_size;
        char    nullable, new_nullable;
        char    defaulted, new_defaulted;
        char    sys_maint_col;
        char    deleted;
        int     col_seq, new_seq;               /* position in base table */
        char    newly_added_fl;                 /* 'Y' if created by "addcol" */
        char    new_definition[100];
        int     ix_seq_arr[MAX_INDEXES +1];     /* position in ALL indexes */
        char    ix_dir_arr[MAX_INDEXES +1];
        char    def_val[1502];                  /* OpenIngres only */
} COL_REC;


/*
| The TBL_REC typedef defines a structure containing all the table
| attributes (except column-names), for use in re-creating the table
| definition and structure (also used for the structure of existing
| secondary indexes).
*/
typedef struct {
        char    name[25];
        char    storage_structure[25];
        char    is_compressed[2];
        char    key_is_compressed[2];
        char    unique_rule[2];
        char    duplicate_rows[2];
        char    is_journalled[2];
        char    grant_all_to_all[2];
        char    grant_ret_to_all[2];
        char    location_str[256];
        int     ifillpct, dfillpct, lfillpct;
        int     minpages, maxpages;
} TBL_REC;



COL_REC         COL_ARR[MAX_COLS_PER_TABLE];
int             n_col_elems;

TBL_REC         TBL_ARR[ MAX_INDEXES +1];
int             n_tbl_elems;
EXEC SQL end declare section;

typedef struct {
        char    view_name[40], view_owner[40];
        int     view_reltid, got_sub_views;
} VIEW_REC;

VIEW_REC        view_arr[MAX_VIEWS];
int             view_arr_elems = 0;



typedef struct {
        char    testing;
        char    commit_after_create;    /* commented-out for safety! */
        char    needs_copyout;
        char    table_exists;
        char    new_table;
        char    copied_out;
        char    abort_no_col;
        char    abort_dup_col;
        char    sort_by_name;
        char    keep_prikey;
        char    keep_altkeys;
        char    keep_rules;
        char    keep_views;
        char    keep_integrities;
        char    keep_permits;
        char    keep_copyoutfile;
        char    keep_journalling;
        char    modify_before_reload;
        char    abort_on_postsql_err;   /* not implemented */
        char    OI_table;       /* is table an OpenIngres table? */
} FLAG_REC;

FLAG_REC        FLAGS;


FILE *cmdf, *psqlf, *logf;

extern struct passwd    *passwd_entry;

/*
| Added some function definitions to get rid of annoying warning messages.
*/
void getuserid(char *);
void panic(char *, char);
int view_compare(VIEW_REC *, VIEW_REC *);
int alpha_compare(COL_REC *, COL_REC *);
void alpha_sort(void);
int newseq_compare(COL_REC *, COL_REC *);
void newseq_sort(void);
int assign_newseq(char *, int);
int find_free_newseq(void);
void set_remaining_newseq(void);
void do_col_sequencing(void);
void build_keycol_str(int, int, char *);
void checkerrors(char *, char);
void connect_as(char *, const char);
void do_sql(char *, char);
void init_flags(void);
void lowercase(char s[]);
void uppercase(char s[]);
void get_col_info();
void get_table_info();
void get_alt_indexes();
void get_loc_str(char *, char *, char *, char *);
int find_col_name(char *);
void build_modify_stmt(char *);
void build_alt_index_stmt(char *, int);
void build_create_col_line(char *, char *, char *, int,
                        char, char, char,
                        char *);
void build_create_stmt(char *);
void build_copy_format_str(char *, char *, char *,
                        char, char, char);
void build_copy_stmt(char *, char *);
char check_ynr(char *);
char check_yn(char *);
void process_cmd(char *);
char line_is_continued(char *);
void get_commands(void);
void save_rules(void);
void save_integrities(void);
void save_views(void);
int lookup_view_elem(char *, char *);
void get_all_views(char *, char *);
void save_permits(char *, char *);
int restore_rules(void);
int restore_views(void);
int restore_integrities(void);
int restore_permits(void);
int do_post_sql_stuff(void);
int restore_stuff(char *);
int got_past_comments;
int got_table_name_keyword;

int main(argc, argv)
  int   argc;
  char *argv[];
{
    int i;
    char line[256], *tmp;
    strcpy(line, argv[1]);

    if (strncmp(argv[1], "-h", 2) == 0) {
      printf("usage:\n");
      printf("Use the the table_name block for as many tables that need to change.\n\n");
      printf("  table_name      <name of table>\n");
      printf("    <following directives>\n");
      printf("  end_table_changes\n\n");
      printf("addcol - specifies column and definition to be added to table\n");
      printf("  addcol <column name>\n\n");
      printf("delcol - specifies column to be dropped from table\n");
      printf("  delcol <column name>\n\n");
      printf("resize - specifies new length for column\n");
      printf("  resize <column name> <length>\n\n");
      printf("resize_numeric - specifies new numeric storage size\n");
      printf("  resize_numeric <column name> <numeric specifier>\n");
      printf("  Valid values for <numeric specifier> are:\n");
      printf("    smallint\n");
      printf("    integer1\n");
      printf("    float4\n");
      printf("    float\n");
      printf("  Numeric sizes of fields can only be changed to a higher order of size.\n\n");
      printf("rename - rename an existing column\n");
      printf("  rename <old column name> <new column name>\n\n");
      printf("sort_col_names - sort the column-names (so HELP table is sorted)\n");
      printf("  sort_col_names = Y|N\n\n");
      printf("colseq - explicitly specifies a column sequence\n");
      printf("  colseq <column name> <column sequence>\n\n");
      printf("The following are used with the create_table directive.\n");
      printf("  withnull        - make a column nullable\n");
      printf("  withdefault     - make a column defaulted\n");
      printf("  notnull         - make a column not nullable\n");
      printf("  notdefault      - make a column not defaulted\n\n");
      printf("create_table - indicates that this is a NEW table (don't try to save)\n");
      printf("The create_table directive can take one of two forms.\n");
      printf("  create_table <table name>\n");
      printf("    addcol <column name>\n");
      printf("      notnull\n");
      printf("      notdefault\n");
      printf("    addcol <column name>\n");
      printf("      notnull\n");
      printf("      notdefault\n");
      printf("    .\n");
      printf("    .\n");
      printf("       OR\n");
      printf("  create_table <table name>\n");
      printf("    addcol <column name> notnull notdefault\n");
      printf("    addcol <column name> notnull notdefault\n");
      printf("    .\n");
      printf("    .\n\n");
      printf("drop_table - drop an existing table, no abort if it doesn't exist\n");
      printf("  drop_table  <table name>\n\n");
      printf("The following keep_* directives are  = Y|N.\n");
      printf("  keep_prikey     - the primary-key/table structure will be kept\n");
      printf("  keep_altkeys    - the secondary key structures will be kept\n");
      printf("  keep_rules      - any rules for the table will be kept\n");
      printf("  keep_views      - any views for the table will be kept\n");
      printf("  keep_integrities- any integrities for the table will be kept\n");
      printf("  keep_permits    - any permits for the table will be kept\n");
      printf("  keep_journalling- option for preserving journaling status of table\n\n");
      printf("post_sql - any Ingres SQL statement (except for 'select')\n");
      printf("  (Note use of line continuation character for addition lines)\n");
      printf("  post_sql   update <table_name> set \\\n");
      printf("               <column1> = <something1> \\\n");
      printf("               <column2> = <something2> \\\n");
      printf("               .                        \\\n");
      printf("               .                               ");
      exit(1);
    }

    printf("%s\n", PROG_NAME);

    if (argc != 3 && argc != 4 && argc != 5) {
        printf("usage: tblmnt  database  script_file  [{-u}as_user] [-fcleanup]\n");
        printf("       tblmnt  -h (For a list of directives used in script_file.\n");
        exit(1);
    }


    memset(COL_ARR, 0, sizeof(COL_ARR));
    memset(TBL_ARR, 0, sizeof(TBL_ARR));

    strcpy(the_db, argv[1]);
    strcpy(the_script, argv[2]);

    if (argc == 5)
        if (strcmp(argv[4], "-fcleanup") == 0)
            CleanUp = 'Y';

    if (argc == 4 || argc == 5)
    {
        if (strncmp(argv[3], "-u", 2) == 0)
            strcpy(the_owner, argv[3]+2);
        else
            strcpy(the_owner, argv[3]);
    }
    else
        getuserid(the_owner);

    /* This doesn't appear to get set anywhere before it is */
    /* tested for and needed.  So set the current user to   */
    /* who is logged in.  This could be different than the  */
    /* table owner if argc 4 or argc 5 is set.              */
    getuserid(curr_user);

    /* on VAX/VMS, cuserid returns an uppercase string - make lowercase */
    lowercase(the_owner);



    /* make sure the user doesn't have autocommit on */
    EXEC SQL set autocommit off;


    /* open the script_file, and find out what table we are dealing with */
    cmdf = fopen(the_script, "r");
    if (cmdf == NULL) {
        printf("cannot open '%s'\n", the_script);
        perror("fopen");
        exit(1);
    }


   connect_as(the_owner, 'Y');

   got_past_comments = FALSE;

   while (!got_past_comments)
   {
      fgets(line, sizeof(line), cmdf);
      /* if it's a comment-line (or line-feed only), skip it */
      if ((line[0] == '#') || (line[0] == '\n'))
         continue;
      else
         got_past_comments = TRUE;
   }

   lowercase(line);
   tmp = strtok(line, " \t");
   if (strcmp(tmp, "table_name"))
      panic("No table specified - first keyword must be table_name", 'Y');
   else {
    tmp = strtok(NULL, " \n");
    strcpy(the_table, tmp);
   }

   while (TRUE)
   {
       printf("\nTABLE: '%s'\n", the_table);
       sprintf(the_datafilename, "%s_%s.dat", the_table, the_owner);


       /* obviously, these names won't work on a system with SHORT filenames! */
       sprintf(the_rulefilename,   "%s_%s.rul", the_table, the_owner);
       sprintf(the_viewfilename,   "%s_%s.vie", the_table, the_owner);
       sprintf(the_integfilename,  "%s_%s.int", the_table, the_owner);
       sprintf(the_permfilename,   "%s_%s.prm", the_table, the_owner);
       sprintf(the_psqlfilename,   "%s_%s.sql", the_table, the_owner);
       sprintf(the_logfilename,    "%s_%s.log", the_table, the_owner);

       logf = fopen(the_logfilename, "w");
       if (logf == NULL) {
           printf("cannot open '%s' for output\n", the_logfilename);
           perror("fopen");
           exit(1);
       }

       /* Not sure why this is ?????   */
       /* curr_user[0] = '\0';         */

       rules_saved = 0;
       integrities_saved = 0;
       views_saved = 0;
       permits_saved = 0;
       post_sql_saved = 0;
       view_arr_elems = 0;
       memset(COL_ARR, 0, sizeof(COL_ARR));
       memset(TBL_ARR, 0, sizeof(TBL_ARR));
       init_flags();
       /* This is initialized when declared but set it here */
       /* again to reset it for additional table changes    */
       post_sql_saved = 0;


       strcpy(TBL_ARR[0].name, the_table);

       printf("Retrieving table information ...\n");
       get_table_info();
       get_col_info();
       get_alt_indexes();

       printf("Parsing the command script ...\n");
       get_commands();


       if ((FLAGS.table_exists != 'Y') && (FLAGS.new_table != 'Y'))
       {
           sprintf(line, "You do not own table '%s' - no changes performed!\n",
                           the_table);
           panic(line, 'N');
       }


       FLAGS.copied_out = 'N';


       /*
       | handle any requests for sorting column sequences...
       */
       do_col_sequencing();

       /* do we need to save the data? */
       if ((FLAGS.table_exists == 'Y') && (FLAGS.needs_copyout == 'Y'))
       {
           if (FLAGS.keep_rules == 'Y')            save_rules();
           if (FLAGS.keep_integrities == 'Y')      save_integrities();

           /* caution - save_permits()  must be called before save_views() */
           if (FLAGS.keep_permits == 'Y')  save_permits(the_table, the_owner);
           if (FLAGS.keep_views == 'Y')            save_views();


           printf("Copying data OUT to '%s' ...\n", the_datafilename);
           build_copy_stmt(STATEMENT, "OUT");
           do_sql(STATEMENT, 'Y');
           FLAGS.copied_out = 'Y';
           EXEC SQL commit;        /* always commit the copy-out! */

           printf("Dropping the table ...\n");
           sprintf(STATEMENT, "drop %s", the_table);
           do_sql(STATEMENT, 'Y');
                                   /* don't even think about commiting this */
       }


       /* do we need to [re]create it? */
       if ((FLAGS.new_table == 'Y') || (FLAGS.copied_out == 'Y'))
       {
           if (FLAGS.copied_out == 'Y')
               printf("Rec");
           else
               printf("C");
           printf("creating the table ...\n");
           build_create_stmt(STATEMENT);
           do_sql(STATEMENT, 'Y');
       }


       /* modify table to base structure BEFORE loading with data? */
       if (FLAGS.modify_before_reload == 'Y') {
           if ((FLAGS.copied_out == 'Y') && (FLAGS.keep_prikey == 'Y'))
           {
               build_modify_stmt(STATEMENT);
               printf("Modifying primary table structure ...\n");
               do_sql(STATEMENT, 'Y');     /* DO abort if can't do the modify */
                                           /* don't commit! */
           }
       }


       if (FLAGS.copied_out == 'Y')
       {
           printf("Reloading table from file ...\n");
           build_copy_stmt(STATEMENT, "IN");
           do_sql(STATEMENT, 'Y');
           printf("<commit>\n");
           EXEC SQL commit;

           if (FLAGS.keep_copyoutfile == 'N')
               remove(the_datafilename);
           else if (CleanUp == 'Y')
               remove(the_datafilename);
       }



       if (FLAGS.modify_before_reload != 'Y') {
           if ((FLAGS.copied_out == 'Y') && (FLAGS.keep_prikey == 'Y'))
           {
               build_modify_stmt(STATEMENT);
               printf("Modifying primary table structure ...\n");
               do_sql(STATEMENT, 'N');     /* DON'T abort if can't do the modify */
               EXEC SQL commit;
           }
       }


       /* do we NEED to re-create secondary keys? */
       if ((FLAGS.copied_out == 'Y') && (FLAGS.keep_altkeys == 'Y'))
       {
           for (i = 1; i < n_tbl_elems; i++)
           {
               printf("Restoring secondary key (#%d) ...\n", i);
               build_alt_index_stmt(STATEMENT, i);
               do_sql(STATEMENT, 'N');
               EXEC SQL commit;
           }
       }


       /* Does journaling need to be turned back on? */
       if (FLAGS.copied_out == 'Y')
       {
           /* need to RESTORE journalling? */
           if ( ((TBL_ARR[0].is_journalled[0] == 'Y')      &&
                (FLAGS.keep_journalling == 'R'))                   ||
                ((TBL_ARR[0].is_journalled[0] == 'C')      &&
                (FLAGS.keep_journalling != 'N'))
              )
           {
               printf("Restoring journaling on table -- it won't be effective\n");
               printf("until the next checkpoint is taken!\n");
               sprintf(STATEMENT, "set journaling on %s", TBL_ARR[0].name);
               do_sql(STATEMENT, 'N');
           }
       }


       if (FLAGS.copied_out == 'Y')
       {
           if (TBL_ARR[0].grant_all_to_all[0] == 'Y')
           {
               sprintf(STATEMENT, " grant all on %s to public ", the_table);
               do_sql(STATEMENT, 'N');
           }

           if (TBL_ARR[0].grant_ret_to_all[0] == 'Y')
           {
               sprintf(STATEMENT, " grant select on %s to public ", the_table);
               do_sql(STATEMENT, 'N');
           }

           EXEC SQL commit;
       }


       /* do we want to re-create the rules/views/... ? */
       if (FLAGS.copied_out == 'Y')
       {
           if (FLAGS.keep_rules == 'Y')
               if (restore_rules() == 0)
                   if (CleanUp == 'Y') remove(the_rulefilename);

           if (FLAGS.keep_views == 'Y')
               if (restore_views() == 0)
                   if (CleanUp == 'Y') remove(the_viewfilename);

           if (FLAGS.keep_integrities == 'Y')
               if (restore_integrities() == 0)
                   if (CleanUp == 'Y') remove(the_integfilename);

           if (FLAGS.keep_permits == 'Y')
               if (restore_permits() == 0)
                   if (CleanUp == 'Y') remove(the_permfilename);
       }

       EXEC SQL commit;



       /* and finally, process any "post_sql" commands ... */
       /* at this time, these are all performed with a pseudo-autocommit */
       if (do_post_sql_stuff() == 0)
           if (CleanUp == 'Y') remove(the_psqlfilename);

       EXEC SQL commit;

       got_table_name_keyword = FALSE;
       while (fgets(line, sizeof(line), cmdf))
       {
          lowercase(line);
          /* read until next table name                         */
          if (!strncmp(line, "table_name", 10))
          {
            got_table_name_keyword = TRUE;
            break;
          }
       }

       if (got_table_name_keyword) {
         /* get table name */
         tmp = strtok(line, " \t");
         tmp = strtok(NULL, " \n");
         if (tmp[0] == '\0')
            panic("No table specified - processing stops", 'Y');
         else {
            strcpy(the_table, tmp);
            continue;
         }
       } else break;
   } /* end while */
    EXEC SQL disconnect;

    return(0);
}

void panic(char *msg, char abort_fl)
{
    if (abort_fl == 'Y') {
        printf("PANIC: %s", msg);
        printf(" - aborting...\n");
        EXEC SQL rollback;
        EXEC SQL disconnect;
        exit(1);
    }
    else
        printf("warning: %s\n", msg);
}


void getuserid(char *user)
{
#ifdef NO_CUSERID
    /* Don't you just love Windows -- yecchhhh! */
    strcpy(user, "");
#else
    cuserid(user);
#endif
}



/* this compares the view_reltid for views */
int view_compare(VIEW_REC *A, VIEW_REC *B)
{
    return (A->view_reltid - B->view_reltid);
}


/*
| This next section of code deals with assigning a column_sequence to
| the columns in the table.  Many thanks to Per Oyvind Overli, for
| coming up with a workable solution (but which I modified quite a
| bit to satisfy some requests to be able to sort columns by name)
*/

        int alpha_compare(COL_REC *A, COL_REC *B)
        {
            char *a, *b;

            a = A->new_name; if (a[0] == '\0') a = A->name;
            b = B->new_name; if (b[0] == '\0') b = B->name;

            return( strcmp(a, b) );
        }

        void alpha_sort()
        {
            qsort(COL_ARR, n_col_elems, sizeof(COL_REC),
                        (int (*)(const void *, const void *)) alpha_compare);
        }


        int newseq_compare(COL_REC *A, COL_REC *B)
        {
            return( A->new_seq - B->new_seq);
        }

        void newseq_sort()
        {
            qsort(COL_ARR, n_col_elems, sizeof(COL_REC),
                        (int (*)(const void *, const void *)) newseq_compare);
        }


        int assign_newseq(char *col_name, int seq)
        {
            int i;

            /* make sure <seq> not already assigned */
            for (i = 0; i < n_col_elems; i++)
                if (COL_ARR[i].new_seq == seq)
                    return(-1);


            /* and assign it... */
            for (i = 0; i < n_col_elems; i++)
            {
                if (strcmp(COL_ARR[i].name, col_name) == 0)
                    COL_ARR[i].new_seq = seq;
            }

            return(1);
        }


        int find_free_newseq()
        {
            int i, x, x_used;

            /* check all possible values (ie 1 .. #cols */
            for (x = 1; x <= n_col_elems; x++)
            {
                x_used = 0;
                for (i = 0; i < n_col_elems; i++)
                {
                    if (COL_ARR[i].deleted == 'Y') continue;
                    if (COL_ARR[i].new_seq == x)
                        x_used = 1;
                }

                if (x_used == 0) return(x);
            }

            return(-1);
        }

        void set_remaining_newseq()
        {
            int i, x;

            for (i = 0; i < n_col_elems; i++)
            {
                if (COL_ARR[i].new_seq == 0)
                    if ((x = find_free_newseq()) != -1)
                        COL_ARR[i].new_seq = x;
            }
        }


        void do_col_sequencing()
        {
            int i;

            /* if sort-by-name specified, do that now [just before copyout] */
            if (FLAGS.sort_by_name == 'Y') alpha_sort();

            /* set the .new_seq field for all column not already set */
            set_remaining_newseq();
            /* and sort by new sequence numbers */
            newseq_sort();
            /*
            | The sorting by name first (if requested), and THEN by the
            | newseq is deliberate - that allows you to specify sorted
            | names for most columns, but maybe one or two that you always
            | want to be first...
            */


            /*
            | If any of the columns truly are in a different order than
            | they were before, we'll need to do a copyout.  We can tell
            | by determining if the original column-sequence (.col_seq)
            | are still in order.
            */
            for (i = 1; i < n_col_elems; i++)
            {
                if (COL_ARR[i].col_seq < COL_ARR[i-1].col_seq)
                    FLAGS.needs_copyout = 'Y';
            }
        }




/*
| The build_key_col() routine builds a comma-delimited string of all the
| columns in a key ...
*/
void build_keycol_str(int IX_NO, int KeyOnly, char *keycol_str)
{
    int   cols_used = 0;
    int   DesiredKeySeq, j;

    keycol_str[0] = '\0';

    /* outer loop: is the ix_seq_arr[] value we are searching for */
    for (DesiredKeySeq = 1; DesiredKeySeq < n_col_elems +1; DesiredKeySeq++)
    {
        /* inner loop: check each COL_ARR row for the DesiredKeySeq */
        for (j = 0; j < n_col_elems; j++)
        {
            /* if col was deleted, don't try to put it back in an index! */
            if (COL_ARR[j].deleted ==  'Y') continue;

            /* also, skip any newly created columns here as well */
            if (COL_ARR[j].newly_added_fl == 'Y') continue;

            if (COL_ARR[j].ix_seq_arr[IX_NO] == DesiredKeySeq)
            {
                if (cols_used > 0) strcat(keycol_str, ",");

                strcat(keycol_str, COL_ARR[j].new_name);

                /* are there descending keys?  I've never seen one! */
                if (COL_ARR[j].ix_dir_arr[IX_NO] == 'D')
                    strcat(keycol_str, " desc");

                cols_used++;
            }
        }
    }


    if (KeyOnly) return;

    /* otherwise, we need the full list of columns for a 2ndary index */
    for (j = 0; j < n_col_elems; j++)
    {
        if (COL_ARR[j].deleted == 'Y') continue;
        if (COL_ARR[j].newly_added_fl == 'Y') continue;
        if (COL_ARR[j].ix_seq_arr[IX_NO] == NONKEY_seq)
        {
            if (cols_used > 0) strcat(keycol_str, ",");
            strcat(keycol_str, COL_ARR[j].new_name);
        }
    }

}



/*
| The checkerrors routine checks for database errors, always performing
| a rollback, but optionally, depending on `abort_on_error' [Y/N] either
| aborting or continuing.
*/
void checkerrors(char *msg, char abort_on_error)
{
    EXEC SQL inquire_ingres(
        :ingrows= rowcount,
        :ingerr = errorno,
        :ingmsg = errortext);

    if (ingerr != 0)
    {
        EXEC SQL rollback;
        printf("%s\n%s\n", msg, ingmsg);
        fprintf(logf,"%s\n%s\n", msg, ingmsg);

        if (abort_on_error == 'Y')
        {
            EXEC SQL disconnect;
            exit(1);
        }
        printf("(continuing anyway)\n\n");
        fprintf(logf,"(continuing anyway)\n\n");
    }

    fflush(logf);
}


void connect_as(char *pv_as_user, const char abort_flag)
{
  EXEC SQL begin declare section;
  char as_user[256];
  EXEC SQL end declare section;
    strcpy (as_user, pv_as_user);
    if (curr_user[0] != '\0')
    {
        EXEC SQL commit;
        EXEC SQL disconnect;
        curr_user[0] = '\0';
    }

    EXEC SQL connect :the_db identified by :as_user
                options = '-f4F30.10', '-f8F50.10';
    checkerrors("Unable to connect to the database", abort_flag);
    if (ingerr == 0)
        strcpy(curr_user, as_user);
    else
    {
        printf("couldn't connect as user [%s]\n", as_user);
        fprintf(logf, "couldn't connect as user [%s]\n", as_user);
    }
}



/*
| The do_sql routine executes an SQL statement.
*/
void do_sql(char *statement, char abort_on_error)
{
    EXEC SQL begin declare section;
        char *stmt_ptr;
    EXEC SQL end declare section;

    stmt_ptr = statement;

    fprintf(logf,"%s\n", stmt_ptr);

    if (FLAGS.testing != 'Y')
    {
        EXEC SQL execute immediate :stmt_ptr;
        checkerrors("Database ERROR", abort_on_error);

        if (ingrows != -1) printf("%d rows\n", ingrows);
        if (ingrows != -1) fprintf(logf,"%d rows\n", ingrows);
    }
}


/*
| Set defaults for FLAGS.
*/
void init_flags()
{
    FLAGS.testing               = 'N';
    FLAGS.commit_after_create   = 'N';
    FLAGS.needs_copyout         = 'N';
    FLAGS.table_exists          = 'N';
    FLAGS.new_table             = 'N';
    FLAGS.copied_out            = 'N';
    FLAGS.abort_no_col          = 'N';
    FLAGS.abort_dup_col         = 'N';
    FLAGS.sort_by_name          = 'N';
    FLAGS.keep_prikey           = 'Y';
    FLAGS.keep_altkeys          = 'Y';
    FLAGS.keep_views            = 'Y';
    FLAGS.keep_rules            = 'Y';
    FLAGS.keep_integrities      = 'Y';
    FLAGS.keep_permits          = 'Y';
    FLAGS.keep_copyoutfile      = 'Y';
    FLAGS.keep_journalling      = 'Y';
    FLAGS.modify_before_reload  = 'N';
    FLAGS.abort_on_postsql_err  = 'N';
    FLAGS.OI_table              = 'N';
}


void lowercase(char s[])
{
    int i;

    for (i = 0; s[i]; i++)
        s[i] = (char)tolower(s[i]);
}


void uppercase(char s[])
{
    int i;

    for (i = 0; s[i]; i++)
        s[i] = (char)toupper(s[i]);
}




/*
| The get_col_info routine retrieves all the info out of IIcolumns
*/
void get_col_info()
{
    EXEC SQL begin declare section;
        char    name[25], type[25], nullable[10], defaulted[10];
        char    sort_dir[10], sys_maint_col[10];
        int     size, colseq, keyseq, found;
        char    col_def_val[1502];
    EXEC SQL end declare section;

    n_col_elems = 0;

    EXEC SQL
        select  squeeze(column_name), squeeze(column_internal_datatype),
                column_length, column_nulls, column_defaults,
                key_sequence, sort_direction, column_system_maintained,
                column_sequence
        into    :name, :type,
                :size, :nullable, :defaulted,
                :keyseq, :sort_dir, :sys_maint_col,
                :colseq
        from    iicolumns
        where   table_name      = :the_table
          and   table_owner     = :the_owner
          and   column_name     != 'tidp'
        order by column_sequence;

    EXEC SQL begin;

        strcpy(COL_ARR[n_col_elems].name, name);
        strcpy(COL_ARR[n_col_elems].new_name, name);
        strcpy(COL_ARR[n_col_elems].type, type);
        COL_ARR[n_col_elems].size = size;
        COL_ARR[n_col_elems].nullable = nullable[0];
        COL_ARR[n_col_elems].new_nullable = nullable[0];
        COL_ARR[n_col_elems].defaulted = defaulted[0];
        COL_ARR[n_col_elems].new_defaulted = defaulted[0];
        COL_ARR[n_col_elems].sys_maint_col = sys_maint_col[0];
        COL_ARR[n_col_elems].newly_added_fl = 'N';
        COL_ARR[n_col_elems].new_definition[0] = '\0';
        COL_ARR[n_col_elems].col_seq = colseq;
        COL_ARR[n_col_elems].ix_seq_arr[0] = keyseq;
        COL_ARR[n_col_elems].ix_dir_arr[0] = sort_dir[0];
        COL_ARR[n_col_elems].def_val[0] = '\0';
        n_col_elems++;

    EXEC SQL end;

    checkerrors("Error selecting from IIcolumns",'Y');
    EXEC SQL commit;


    /* check to see if there are specific default values (OpenIngres only!) */
    if (FLAGS.OI_table == 'Y')
    {
        EXEC SQL REPEATED
        select  squeeze(column_name), squeeze(column_default_val)
        into    :name, :col_def_val
        from    iicolumns
        where   table_name = :the_table
          and   table_owner = :the_owner
          and   column_has_default = 'Y';

        EXEC SQL begin;

                found = find_col_name(name);
                strcpy(COL_ARR[found].def_val, col_def_val);

        EXEC SQL end;

        checkerrors("Error selecting from IIcolumns", 'Y');
        EXEC SQL commit;
    }
}


void get_table_info()
{
    EXEC SQL begin declare section;
        char    table_type[40], location_name[40], multi_locations[10];
        char    table_version[10];
    EXEC SQL end declare section;

    EXEC SQL
        select  squeeze(storage_structure), squeeze(is_compressed),
                squeeze(key_is_compressed), squeeze(unique_rule),
                squeeze(location_name), squeeze(multi_locations),
                table_ifillpct, table_dfillpct, table_lfillpct,
                table_minpages, table_maxpages, table_type,
                squeeze(duplicate_rows), squeeze(is_journalled),
                squeeze(all_to_all), squeeze(ret_to_all),
                squeeze(table_version)
        into    :TBL_ARR[0].storage_structure,  :TBL_ARR[0].is_compressed,
                :TBL_ARR[0].key_is_compressed,  :TBL_ARR[0].unique_rule,
                :location_name,                 :multi_locations,
                :TBL_ARR[0].ifillpct,           :TBL_ARR[0].dfillpct,
                :TBL_ARR[0].lfillpct,           :TBL_ARR[0].minpages,
                :TBL_ARR[0].maxpages,           :table_type,
                :TBL_ARR[0].duplicate_rows,     :TBL_ARR[0].is_journalled,
                :TBL_ARR[0].grant_all_to_all,   :TBL_ARR[0].grant_ret_to_all,
                :table_version
        from    iitables
        where   table_name      = :the_table
          and   table_owner     = :the_owner;

    checkerrors("Error selecting from IItables",'Y');
    if (ingrows == 0)
    {
        printf("Table '%s' does not exist!\n", the_table);
        return;
    }

    FLAGS.table_exists = 'Y';

    if (table_type[0] != 'T')
    {
        printf("'%s' not a TABLE (type = '%s')\n", the_table, table_type);
        exit(1);
    }

    if (strncmp(table_version, "OPING", 5) == 0)
        FLAGS.OI_table = 'Y';
    else
        FLAGS.OI_table = 'N';


    get_loc_str(TBL_ARR[0].location_str,
                        location_name, multi_locations, the_table);

    n_tbl_elems = 1;
}




/*
| This brain-dead little routine loads the index-structures for each of
| the table's secondary indexes.
*/
void get_alt_indexes()
{
    EXEC SQL begin declare section;
        char    index_name[25], storage_structure[25], is_compressed[2];
        char    unique_rule[2], location_name[40], multi_locations[2];
        char    col_name[25], sort_dir[10], key_is_compressed[2];
        int     ifillpct, dfillpct, lfillpct, minpages, maxpages, key_seq;
        int     found, LastKeySeq;
    EXEC SQL end declare section;

    EXEC SQL declare key_csr cursor for
        select  squeeze(i.index_name), squeeze(i.storage_structure),
                squeeze(i.is_compressed), squeeze(i.unique_rule),
                squeeze(t.location_name), squeeze(t.multi_locations),
                t.table_ifillpct, t.table_dfillpct, t.table_lfillpct,
                t.table_minpages, t.table_maxpages,
                squeeze(i.key_is_compressed)
        from    iiindexes i, iitables t
        where   i.base_name     = :the_table
          and   i.base_owner    = :the_owner
          and   i.index_owner   = :the_owner
          and   t.table_name    = i.index_name
          and   t.table_owner   = :the_owner ;

    EXEC SQL open key_csr for readonly;

    while (sqlca.sqlcode == 0)
    {
        EXEC SQL fetch key_csr into
                :index_name,    :storage_structure,
                :is_compressed, :unique_rule,
                :location_name, :multi_locations,
                :ifillpct,      :dfillpct,
                :lfillpct,      :minpages,
                :maxpages;

        if (sqlca.sqlcode != 0) break;

        /*
        | FIXME -- it would be a lot nicer to select this info directly
        |          into the TBL_ARR[ n ] structure!
        */
        strcpy(TBL_ARR[n_tbl_elems].name, index_name);
        strcpy(TBL_ARR[n_tbl_elems].storage_structure, storage_structure);
        strcpy(TBL_ARR[n_tbl_elems].is_compressed, is_compressed);
        strcpy(TBL_ARR[n_tbl_elems].key_is_compressed, key_is_compressed);
        strcpy(TBL_ARR[n_tbl_elems].unique_rule, unique_rule);
        get_loc_str(TBL_ARR[n_tbl_elems].location_str,
                                location_name, multi_locations, index_name);
        TBL_ARR[n_tbl_elems].ifillpct = ifillpct;
        TBL_ARR[n_tbl_elems].dfillpct = dfillpct;
        TBL_ARR[n_tbl_elems].lfillpct = lfillpct;
        TBL_ARR[n_tbl_elems].minpages = minpages;
        TBL_ARR[n_tbl_elems].maxpages = maxpages;


        /*
        | scan the list of columns for the current index, and keep track
        | of it's column-sequence.
        */
        LastKeySeq = 0;
        EXEC SQL REPEATED
                select  squeeze(column_name), key_sequence,
                        squeeze(sort_direction)
                into    :col_name, :key_seq, :sort_dir
                from    IIindex_columns
                where   index_name = :index_name
                  and   index_owner = :the_owner
                order by key_sequence;

        EXEC SQL begin;

                found = find_col_name(col_name);
                COL_ARR[found].ix_seq_arr[n_tbl_elems] = key_seq;
                COL_ARR[found].ix_dir_arr[n_tbl_elems] = sort_dir[0];

        EXEC SQL end;

        checkerrors("Error selecting from IIindex_columns",'Y');
        sqlca.sqlcode = 0;


        /* don't forget the "DATA" columns for the 2ndary index */
        EXEC SQL REPEATED
                select  squeeze(column_name)
                into    :col_name
                from    IIcolumns
                where   table_name = :index_name
                  and   table_owner = :the_owner
                  and   key_sequence = 0;

        EXEC SQL begin;

            found = find_col_name(col_name);
            COL_ARR[found].ix_seq_arr[n_tbl_elems] = NONKEY_seq;

        EXEC SQL end;
        checkerrors("Error selecting from IIcolumns", 'Y');
        sqlca.sqlcode = 0;

        n_tbl_elems++;
    }
    checkerrors("Error fetching from IIindexes/IItables",'Y');
    EXEC SQL commit;
}



void get_loc_str(LocStr, loc_name, multi_locs, obj_name)
  EXEC SQL begin declare section;
  char *LocStr, *loc_name, *multi_locs, *obj_name;
  EXEC SQL end declare section;
{
    EXEC SQL begin declare section;
        char    location[40];
    EXEC SQL end declare section;

    sprintf(LocStr, "location = (%s", loc_name);

    if (multi_locs[0] != 'N')
    {
        EXEC SQL REPEATED
        select  location_name
        into    :location
        from    iimulti_locations
        where   table_name = :obj_name
          and   table_owner = :the_owner;

        EXEC SQL begin;

                strcat(LocStr, ", ");
                strcat(LocStr, location);

        EXEC SQL end;
        checkerrors("Error selecting from IImulti_locations (get_loc_str)",'Y');
        sqlca.sqlcode = 0;
    }

    strcat(LocStr, ") ");
}


int find_col_name(char *name)
{
    int i;

    for (i = 0; i < n_col_elems; i++)
    {
        if (strcmp(COL_ARR[i].name, name) == 0)
            return (i);
    }

    /* not found */
    return (-1);
}



/*
| The build_modify_stmt routine builds the "modify ... to ..." statement
| such that that it is identical to the previous table-structure.
*/
void build_modify_stmt(char *ModStr)
{
    char  keycol_str[512];
    char  t_comp1[10], t_comp2[10];


    if (strcmp(TBL_ARR[0].storage_structure, "HEAP") == 0)
    {
        sprintf(ModStr, "modify %s to HEAP", the_table);
        if (TBL_ARR[0].is_compressed[0] == 'Y')
            sprintf(ModStr, "%s with compression = (data)", ModStr);
    }
    else
    {
        sprintf(ModStr, "modify %s to %s ",
                        the_table, TBL_ARR[0].storage_structure);

        if (TBL_ARR[0].unique_rule[0] == 'U')
            strcat(ModStr, "UNIQUE ");

        build_keycol_str(0, 1, keycol_str);
        sprintf(ModStr, "%s on %s with ", ModStr, keycol_str);

        if (strcmp(TBL_ARR[0].storage_structure, "BTREE") == 0)
            sprintf(ModStr, "%s nonleaffill = %d, ", ModStr,
                                                        TBL_ARR[0].ifillpct);

        if ((strcmp(TBL_ARR[0].storage_structure, "BTREE") == 0)        ||
            (strcmp(TBL_ARR[0].storage_structure, "HASH")  == 0)        ||
            (strcmp(TBL_ARR[0].storage_structure, "ISAM")  == 0))
                sprintf(ModStr, "%s fillfactor = %d, ", ModStr,
                                                        TBL_ARR[0].dfillpct);

        if (strcmp(TBL_ARR[0].storage_structure, "BTREE") == 0)
            sprintf(ModStr, "%s leaffill = %d, ", ModStr,
                                                        TBL_ARR[0].lfillpct);

        if (strcmp(TBL_ARR[0].storage_structure, "HASH") == 0)
        {
            if (TBL_ARR[0].minpages > 0)
                sprintf(ModStr, "%s minpages = %d, ", ModStr,
                                                        TBL_ARR[0].minpages);
            if ((TBL_ARR[0].maxpages > 0) &&
                (TBL_ARR[0].maxpages >= TBL_ARR[0].minpages))
                sprintf(ModStr, "%s maxpages = %d, ", ModStr,
                                                        TBL_ARR[0].maxpages);
        }

        if (TBL_ARR[0].is_compressed[0] != 'Y')
            strcpy(t_comp1, "NO");
        else
            t_comp1[0] = '\0';

        if (TBL_ARR[0].key_is_compressed[0] != 'Y')
            strcpy(t_comp2, "NO");
        else
            t_comp2[0] = '\0';


        sprintf(ModStr, "%s compression = (%sdata, %skey)",
                        ModStr, t_comp1, t_comp2);
    }
}



void build_alt_index_stmt(char *str, int ix_no)
{
    char  key_keycol_str[2048], keycol_str[2048];
    char  with_str[5000], t_comp1[3], t_comp2[3];

    strcpy(str, "create ");

    if (TBL_ARR[ix_no].unique_rule[0] == 'U')
        strcat(str, "UNIQUE ");

    strcat(str, "index ");
    strcat(str, TBL_ARR[ix_no].name);
    strcat(str, " on ");
    strcat(str, the_table);
    strcat(str, " ( ");
    build_keycol_str(ix_no, 1, key_keycol_str);
    build_keycol_str(ix_no, 0, keycol_str);
    strcat(str, keycol_str);
    strcat(str, " )");

    sprintf(with_str, "with structure = %s,", TBL_ARR[ix_no].storage_structure);
    if (strcmp(TBL_ARR[ix_no].storage_structure, "BTREE") == 0)
        sprintf(with_str, "%s nonleaffill = %d,", with_str,
                                                TBL_ARR[ix_no].ifillpct);

    if ((strcmp(TBL_ARR[ix_no].storage_structure, "BTREE")  == 0)       ||
        (strcmp(TBL_ARR[ix_no].storage_structure, "HASH")   == 0)       ||
        (strcmp(TBL_ARR[ix_no].storage_structure, "ISAM")   == 0))
            sprintf(with_str, "%s fillfactor = %d, ", with_str,
                                                TBL_ARR[ix_no].dfillpct);

    if (strcmp(TBL_ARR[ix_no].storage_structure, "BTREE") == 0)
            sprintf(with_str, "%s leaffill = %d, ", with_str,
                                                TBL_ARR[ix_no].lfillpct);

    if (strcmp(TBL_ARR[ix_no].storage_structure, "HASH") == 0)
    {
        if (TBL_ARR[ix_no].minpages > 0)
                sprintf(with_str, "%s minpages = %d, ", with_str,
                                                TBL_ARR[ix_no].minpages);
        if ((TBL_ARR[ix_no].maxpages > 0) &&
            (TBL_ARR[ix_no].maxpages >= TBL_ARR[ix_no].minpages))
                sprintf(with_str, "%s maxpages = %d, ", with_str,
                                                TBL_ARR[ix_no].maxpages);
    }

    if (TBL_ARR[ix_no].is_compressed[0] != 'Y')
        strcpy(t_comp1, "NO");
    else
        t_comp1[0] = '\0';

    if (TBL_ARR[ix_no].key_is_compressed[0] != 'Y')
        strcpy(t_comp2, "NO");
    else
        t_comp2[0] = '\0';

    sprintf(with_str, "%s compression = (%sdata, %skey),",
                        with_str, t_comp1, t_comp2);

    sprintf(with_str, "%s key=(%s),", with_str, key_keycol_str);

    strcat(with_str, TBL_ARR[ix_no].location_str);
    sprintf(str, "%s %s", str, with_str);
}




/*
| The build_create_col_line routine builds column-description for use in
| the "create table ..." statement.  If and when Ingres gets any new
| datatypes, this section of code will need to be tweaked.
*/
void build_create_col_line(char *line, char *name, char *type, int size,
                        char nullable, char defaulted, char sysmaintcol,
                        char *default_value)
{
    sprintf(line, "%s %s", name, type);

    /* is there a SIZE associated with the datatype? */
    if ((strcmp(type,   "INTEGER")      == 0)   ||
        (strcmp(type,   "FLOAT")        == 0))
    {
        sprintf(line,"%s%d", line, size);
    }
    else
    if ((strcmp(type,   "VARCHAR")      == 0)   ||
        (strcmp(type,   "CHAR")         == 0)   ||
        (strcmp(type,   "TEXT")         == 0)   ||
        (strcmp(type,   "C")            == 0)   ||
        (strcmp(type,   "BYTE")         == 0)   ||      /* BYTE -> OpenIngres */
        (strcmp(type,   "BYTE VARYING") == 0))
    {
        sprintf(line,"%s(%d)", line, size);
    }
    if (nullable == 'Y')
    {
        strcat(line, " with null");
    }
    else
    {
        strcat(line, " not null");
        if (defaulted == 'Y')
        {
            strcat(line, " with default ");
            if (FLAGS.OI_table == 'Y')
            {
                strcat(line, default_value);
            }
        }
        else
            strcat(line, " not default");
    }

    if (sysmaintcol == 'Y')
        strcat(line, " with system_maintained");

}



void build_create_stmt(char *str)
{
    int   i, lines_in_stmt = 0;
    char  line[256], t_jnl[10], t_dups[10];

    sprintf(str, "CREATE TABLE %s (", the_table);
    for (i = 0; i < n_col_elems; i++)
    {
        /* skip any dropped columns */
        if (COL_ARR[i].deleted == 'Y') continue;

/*
        if (COL_ARR[i].newly_added_fl != 'Y')
        {
*/
            build_create_col_line( line,
                        COL_ARR[i].new_name,
                        COL_ARR[i].type,
                        COL_ARR[i].size,
                        COL_ARR[i].new_nullable,
                        COL_ARR[i].new_defaulted,
                        COL_ARR[i].sys_maint_col,
                        COL_ARR[i].def_val);
/*
        }
        else
        {
            sprintf(line, "%s %s", COL_ARR[i].name, COL_ARR[i].new_definition);
        }
*/
        if (lines_in_stmt > 0) strcat(str, ",");
        sprintf(str, "%s\n\t%s", str, line);
        lines_in_stmt++;
    }
    strcat(str, ")\n");

    /*
    | for NEW tables, the user will need to specify these things in a
    | post_sql MODIFY statement.  There won't be any "with" clause.
    */

    /*
    | However, if this is an existing table, we'll put some appropriate
    | info into the with-clause for this create statement...
    */
    if (FLAGS.new_table != 'Y')
    {
        /* and finally, append the necessary table-attributes ... */
        if (TBL_ARR[0].duplicate_rows[0] == 'U')
            strcpy(t_dups, "NO");
        else
            t_dups[0] = '\0';

        if ((TBL_ARR[0].is_journalled[0] != 'Y')        ||
            (FLAGS.keep_journalling != 'Y'))
                strcpy(t_jnl, "NO");
        else
            t_jnl[0] = '\0';

        sprintf(str, "%s with %sduplicates, %sjournaling, %s",
                str, t_dups, t_jnl, TBL_ARR[0].location_str);
    }
}



void build_copy_format_str(char *in_or_out, char *format, char *type,
                        char nullable, char new_nullable, char sysmaint)
{
    /* beats me - this is what COPYDB does for these sorts of columns! */
    if (sysmaint == 'Y')
    {
        strcpy(format, "d0tab");
        return;
    }

    if ((strcmp(type, "INTEGER")        == 0)   ||
        (strcmp(type, "MONEY")          == 0)   ||
        (strcmp(type, "DECIMAL")        == 0)   ||
        (strcmp(type, "FLOAT")          == 0)   ||
        (strcmp(type, "DATE")           == 0))
    {
                strcpy(format, "c0tab ");
    }
    else
    {
                strcpy(format, "varchar(0)tab ");
    }

    /* are we making a nullable column non-nullable? */
    if ((nullable == 'Y') && (new_nullable == 'N'))
    {
      if (strcmp(in_or_out, "IN") != 0)
      {
        if ((strcmp(type, "INTEGER")    == 0)   ||
            (strcmp(type, "DECIMAL")    == 0)   ||
            (strcmp(type, "MONEY")      == 0)   ||
            (strcmp(type, "FLOAT")      == 0))
                strcat(format, " with null('0')");
        else
                strcat(format, " with null('')");
      }
    }
    else
    {
        if (nullable == 'Y')
            strcat(format, " with null(']^NULL^[')");
    }
}



void build_copy_stmt(char *str, char *in_or_out)
{
    int   i;
    char  copyformat[256], the_name[40];


    sprintf(str, "copy %s (\n", the_table);
    for (i = 0; i < n_col_elems; i++)
    {
        /* was the column deleted? */
        if (COL_ARR[i].deleted == 'Y') continue;

        /* was the column newly added? */
        if (COL_ARR[i].newly_added_fl == 'Y') continue;

        build_copy_format_str(in_or_out, copyformat,
                COL_ARR[i].type, COL_ARR[i].nullable, COL_ARR[i].new_nullable,
                                                COL_ARR[i].sys_maint_col);

        if (strcmp(in_or_out, "IN") == 0)
            strcpy(the_name, COL_ARR[i].new_name);
        else
            strcpy(the_name, COL_ARR[i].name);

        sprintf(str,"%s  %s = %s,\n", str, the_name, copyformat);
    }

    strcat(str, "nl = d1)\n");
    if (strcmp(in_or_out,"IN") == 0)
        sprintf(str, "%s FROM '%s'", str, the_datafilename);
    else
        sprintf(str, "%s INTO '%s'", str, the_datafilename);
}


char check_ynr(char *str)
{
    char ch;

    if (str == NULL)
        ch = ' ';
    else
        ch = (char)toupper(str[0]);

    if (ch == 'Y' || ch == 'N' || ch == 'R')
        return (ch);
    else
    {
        sprintf(msg,"Bad flag-value [%c] - expecting Y, N, or R\n", ch);
        panic(msg, 'Y');
    }
    return ('X');
}


char check_yn(char *str)
{
    char ch;

    if (str == NULL)
        ch = ' ';
    else
        ch = (char)toupper(str[0]);

    if (ch == 'Y' || ch == 'N')
        return (ch);
    else
    {
        sprintf(msg,"Bad flag-value [%c] - expecting Y or N\n", ch);
        panic(msg, 'Y');
    }
    return ('X');
}



/*
| FIXME - the process_cmd routine is done by brute-force!
*/
void process_cmd(char *cmd_line)
{
    char *command, *col_name, *new_col_name, *tmp, *colseq_str;
    char new_type[33];
    int  new_size, found, new_found, colseq;

    /* This actually allows comments that don't start on     */
    /* the first character position of a line.               */
    if ((command = strtok(cmd_line, " \t")) == NULL) return;
    lowercase(command);

    /* let a "#" be a comment - if first non-whitespace char */
    if (command[0] == '#') return;

   /* If we get here and a command == 'table_name' then an  */
   /* end_table_changes wasn't found before the next table  */
   /* info in the file.                                     */
   if (strcmp(command, "table_name") == 0) {
      sprintf(msg, "bad file format -\n\tno end_table_changes for table '%s'", the_table);
      panic(msg, 'Y');
   }

    /* check for processing directives */
    if (strcmp(command, "abort_dup_col") == 0)
        FLAGS.abort_dup_col     = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "abort_no_col") == 0)
        FLAGS.abort_no_col      = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "commit_after_create") == 0)
        FLAGS.commit_after_create = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "sort_col_names") == 0)
        FLAGS.sort_by_name      = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "keep_prikey") == 0)
        FLAGS.keep_prikey       = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "keep_altkeys") == 0)
        FLAGS.keep_altkeys      = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "keep_rules") == 0)
        FLAGS.keep_rules        = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "keep_views") == 0)
        FLAGS.keep_views        = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "keep_integrities") == 0)
        FLAGS.keep_integrities  = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "keep_permits") == 0)
        FLAGS.keep_permits      = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "modify_before_reload") == 0)
        FLAGS.modify_before_reload      = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "keep_copyoutfile") == 0)
        FLAGS.keep_copyoutfile  = check_yn(strtok(NULL, " =\t"));

    else if (strcmp(command, "testing") == 0)
        FLAGS.testing           = check_yn(strtok(NULL, " =\t"));

    else if ((strcmp(command, "keep_journalling") == 0) ||
            (strcmp(command, "keep_journaling") == 0))
        FLAGS.keep_journalling = check_ynr(strtok(NULL, " =\t"));
        /*
        | yep, talk about a kludge -- completely new procedure
        | check_ynr() -- merely to handle new 'R' flag-value!
        */

    else if (strcmp(command, "create_table") == 0)
    {
        if (FLAGS.table_exists == 'Y') panic("Table already exists!", 'Y');
        FLAGS.new_table = 'Y';
    }

    else if (strcmp(command, "drop_table") == 0)
    {
        if (FLAGS.table_exists == 'Y') {
           printf("Dropping the table ...\n");
           sprintf(STATEMENT, "drop %s", the_table);
           do_sql(STATEMENT, 'N');
           FLAGS.table_exists = 'N';
           init_flags();
           /* This is initialized when declared but set it here */
           /* again to reset it for additional table changes    */
           post_sql_saved = 0;
           strcpy(TBL_ARR[0].name, the_table);
           printf("Resetting table information ...\n");
           get_table_info();
           get_col_info();
           get_alt_indexes();
        }
    }

    else if (strcmp(command, "addcol") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad addcol statement - no column-name specified", 'Y');
        lowercase(col_name);

        found = find_col_name(col_name);
        if (found == -1)
        {
            strcpy(COL_ARR[n_col_elems].name, col_name);
            strcpy(COL_ARR[n_col_elems].new_name, col_name);
            if ((tmp = strtok(NULL, " \t")) == NULL)
                panic("bad addcol statement - no column-def specified", 'Y');
            COL_ARR[n_col_elems].newly_added_fl = 'Y';
            strcpy(COL_ARR[n_col_elems].new_definition, tmp);
            strcpy(COL_ARR[n_col_elems].type, tmp);
            FLAGS.needs_copyout = 'Y';
            if ((tmp = strtok(NULL, " \t")) == NULL)
                panic("bad addcol statement - no nullable specification", 'Y');
            lowercase(tmp);
            if (!strcmp(tmp,"notnull")) {
               COL_ARR[n_col_elems].new_nullable = 'N';
               if ((tmp = strtok(NULL, " \t")) == NULL)
                  panic("bad addcol statement - no default specification", 'Y');
               lowercase(tmp);
               if (!strcmp(tmp,"notdefault"))
                  COL_ARR[n_col_elems].new_defaulted = 'N';
               else
                  COL_ARR[n_col_elems].new_defaulted = 'Y';
            }
            else
               COL_ARR[n_col_elems].new_nullable = 'Y';
            n_col_elems++;
            printf("adding new column '%s'\n", col_name);
        }
        else
        {
            sprintf(msg,"addcol %s - column already exists", col_name);
            panic(msg, FLAGS.abort_dup_col);
        }
    }

    else if (strcmp(command, "delcol") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad delcol statement - no column-name specified", 'Y');
        lowercase(col_name);

        found = find_col_name(col_name);
        if (found >= 0)
        {
            COL_ARR[found].deleted = 'Y';
            FLAGS.needs_copyout = 'Y';
            printf("deleting column '%s'\n", col_name);
        }
        else
        {
            sprintf(msg, "delcol %s - column does not exist", col_name);
            panic(msg, FLAGS.abort_no_col);
        }
    }

    else if (strcmp(command, "resize") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad resize statement - no column-name specified", 'Y');
        lowercase(col_name);

        found = find_col_name(col_name);
        if (found >= 0)
        {
            if ((tmp = strtok(NULL, " \t")) == NULL)
                panic("Invalid resize statement - no length specified", 'Y');
            new_size = atoi(tmp);
            if (new_size == 0)
                panic("Invalid resize statement - bad length specified", 'Y');

            /* make sure the size differs! */
            if (COL_ARR[found].size != new_size)
            {
                printf("changing size of '%s' from %d to %d\n",
                        col_name, COL_ARR[found].size, new_size);
                COL_ARR[found].size = new_size;
                FLAGS.needs_copyout = 'Y';
            }
        }
        else
            panic("resize - column doesn't exist", FLAGS.abort_no_col);
    }

    else if (strcmp(command, "resize_numeric") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad resize statement - no column-name specified", 'Y');
        lowercase(col_name);

        found = find_col_name(col_name);
        if (found >= 0)
        {
            if ((tmp = strtok(NULL, " \t")) == NULL)
                panic("Invalid resize statement - no length specified", 'Y');

            lowercase(tmp);
            if ( (!strncmp(COL_ARR[found].type, "C", 1) OR
                  !strncmp(COL_ARR[found].type, "V", 1) OR
                  !strncmp(COL_ARR[found].type, "D", 1) OR
                  !strncmp(COL_ARR[found].type, "M", 1)) AND
                 (!strncmp(tmp, "i", 1) OR
                  !strncmp(tmp, "s", 1) OR
                  !strncmp(tmp, "f", 1)) )
               panic("Invalid resize statement - conversion of Character,\n\tVarChar, Date or Money types to numeric is unsupported", 'Y');

printf("requested - tmp: %s\n", tmp);


            new_type[0] = '\0';
            strcpy(new_type, COL_ARR[found].type);
printf("new_type: %s\n", new_type);
printf("found size: %d\n", COL_ARR[found].size);
            if (!strncmp(COL_ARR[found].type, "I", 1) AND
                COL_ARR[found].size == 1)
            {
               if (!strcmp(tmp, "smallint"))
                  new_size = 2;
               else if (!strcmp(tmp, "integer1"))
                  new_size = 4;
               else if (!strcmp(tmp, "float4"))
               {
                  new_size = 4;
                  strcpy(new_type, "FLOAT");
               }
               else if (!strcmp(tmp, "float"))
               {
                  new_size = 8;
                  strcpy(new_type, "FLOAT");
               }
               else
                  panic("Invalid resize statement - bad length specified", 'Y');
            }
            else if (!strncmp(COL_ARR[found].type, "I", 1) AND
                     COL_ARR[found].size == 2)
            {
               if (!strcmp(tmp, "integer"))
                  new_size = 4;
               else if (!strcmp(tmp, "float4"))
               {
                  new_size = 4;
                  strcpy(new_type, "FLOAT");
               }
               else if (!strcmp(tmp, "float"))
               {
                  new_size = 8;
                  strcpy(new_type, "FLOAT");
               }
               else
                  panic("Invalid resize statement - bad length specified", 'Y');
            }
            else if (!strncmp(COL_ARR[found].type, "I", 1) AND
                     COL_ARR[found].size == 4)
            {
               if (!strcmp(tmp, "float4"))
               {
                  new_size = 4;
                  strcpy(new_type, "FLOAT");
               }
               else if (!strcmp(tmp, "float"))
               {
                  new_size = 8;
                  strcpy(new_type, "FLOAT");
               }
               else
                  panic("Invalid resize statement - bad length specified", 'Y');
            }
            else if (!strncmp(COL_ARR[found].type, "F", 1) AND
                     COL_ARR[found].size == 4)
            {
               if (!strcmp(tmp, "float"))
                  new_size = 8;
               else
                  panic("Invalid resize statement - bad length specified", 'Y');
            }

            else
               panic("Invalid resize statement - bad length specified", 'Y');
        }

        printf("changing size of '%s' from %s%d to %s%d\n",
                col_name, COL_ARR[found].type,COL_ARR[found].size,
                new_type, new_size);

        COL_ARR[found].size = new_size;
        strcpy(COL_ARR[found].type, new_type);
        FLAGS.needs_copyout = 'Y';
    }
    else if (strcmp(command, "rename") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("rename - no column specified", 'Y');
        lowercase(col_name);
        found = find_col_name(col_name);
        if (found >= 0)
        {
            if ((new_col_name = strtok(NULL, "")) == NULL)
                panic("Invalid rename statement - no new name specified", 'Y');
            lowercase(new_col_name);

            new_found = find_col_name(new_col_name);
            if (new_found >= 0)
            {
                panic("Invalid rename statement - new name already exists",
                                FLAGS.abort_dup_col);
            }
            else
            {
                strcpy(COL_ARR[found].new_name, new_col_name);
                printf("renaming %s to %s\n", col_name, new_col_name);
                FLAGS.needs_copyout = 'Y';
            }
        }
        else
            panic("rename - column doesn't exist", FLAGS.abort_no_col);
    }

    else if (strcmp(command, "notnull") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad notnull statement - no column specified", 'Y');
        lowercase(col_name);

        found = find_col_name(col_name);
        if (found < 0)
        {
            panic("notnull - column doesn't exist", FLAGS.abort_no_col);
        }
        else
        {
            /* only do this if the column IS nullable! */
            if (COL_ARR[found].nullable == 'Y')
            {
                COL_ARR[found].new_nullable = 'N';
                FLAGS.needs_copyout = 'Y';
                printf("making '%s' not nullable\n", col_name);
            }
        }
    }

    else if (strcmp(command, "withnull") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad withnull statement - no column specified", 'Y');
        lowercase(col_name);

        found = find_col_name(col_name);
        if (found < 0)
            panic("withnull - column doesn't exist", FLAGS.abort_no_col);
        else
        {
            /* only do this if the column isn't already WITH NULL */
            if (COL_ARR[found].nullable == 'N')
            {
                COL_ARR[found].new_nullable = 'Y';
                FLAGS.needs_copyout = 'Y';
                printf("making '%s' nullable\n", col_name);
            }
        }
    }

    else if (strcmp(command, "withdefault") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad withdefault statement - no column specified", 'Y');
        lowercase(col_name);

        found = find_col_name(col_name);
        if (found < 0)
            panic("withdefault - column doesn't exist",FLAGS.abort_no_col);
        else
        {
            /* only do this if the column is NOT DEFAULT */
            if (COL_ARR[found].defaulted == 'N')
            {
                COL_ARR[found].new_defaulted = 'Y';
                FLAGS.needs_copyout = 'Y';
                printf("making '%s' defaulted\n", col_name);
            }
        }
    }

    else if (strcmp(command, "notdefault") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad notdefault statement - no column specified", 'Y');
        lowercase(col_name);

        found = find_col_name(col_name);
        if (found < 0)
            panic("notdefault - column doesn't exist",FLAGS.abort_no_col);
        else
        {
            /* only do this if the column is DEFAULTed */
            if (COL_ARR[found].defaulted == 'Y')
            {
                COL_ARR[found].new_defaulted = 'N';
                FLAGS.needs_copyout = 'Y';
                printf("making '%s' not-defaulted\n", col_name);
            }
        }
    }

    else if (strcmp(command, "colseq") == 0)
    {
        if ((col_name = strtok(NULL, " \t")) == NULL)
            panic("bad colseq statement - no column specified", 'Y');

        lowercase(col_name);
        found = find_col_name(col_name);

        if (found < 0)
            panic("colseq - column doesn't exist", FLAGS.abort_no_col);
        else
        {
            if ((colseq_str = strtok(NULL, " \t")) == NULL)
                panic("bad colseq statement - no sequence number", 'Y');

            colseq = atoi(colseq_str);
            if (colseq < 1) panic("bad colseq statement - colseq < 1", 'Y');
            printf("colseq '%s' '%d'\n", col_name,colseq);
            assign_newseq(col_name, colseq);
        }
    }

    else if (strcmp(command, "post_sql") == 0)
    {
        /* no checking for "type" of sql statement! */
        /* just write it out to a file - will be processed LATER */

        if ((tmp = strtok(NULL, "")) == NULL)
            panic("post_sql - no statement defined", 'Y');

        fprintf(psqlf, "%s\t%s\n", the_owner, tmp);
        post_sql_saved++;
    }

    else if (command[0] != '\0')
    {
        sprintf(msg, "Invalid command: [%s]", cmd_line);
        panic(msg, 'Y');
    }
}


char line_is_continued(char *line)
{
    char *p1, *p2;

    p1 = strrchr(line, '\\');
    if (p1 == NULL) return('N');

    for (p2 = p1, p2++; *p2; p2++)
        if ((*p2 != ' ') && (*p2 != '\t'))
            return('N');

    *p1 = ' ';
    return('Y');
}


void get_commands()
{
    char  line[1024], command[8096];
    char  prev_line_continued, line_continued;
    int   last;

    /*
    | open the post-sql file, to save any "post_sql" statements until AFTER
    | all other table-changes have been done.
    */
    psqlf = fopen(the_psqlfilename, "w");
    if (psqlf == NULL)
        panic("couldn't open the post-sql file for output", 'Y');


    /*
    | The cmdf file was opened in main() - to retrieve the table-name from
    | the first line...
    */
    prev_line_continued = 'N';
    while (fgets(line, sizeof(line), cmdf))
    {
        /* if it's a comment-line (or line-feed only), skip it */
        if ((line[0] == '#') || (line[0] == '\n')) continue;
        if (!strncmp(line, "end_table_changes", 17)) break;

        last = strlen(line) -1;
        if (line[last] == '\n') line[last--] = '\0';

        line_continued = line_is_continued(line);

        if (prev_line_continued == 'Y')
            strcat(command, line);
        else
            strcpy(command, line);

        if (line_continued == 'N')
            process_cmd(command);

        prev_line_continued = line_continued;
    }

    fclose(psqlf);
}



void save_rules()
{
    EXEC SQL begin declare section;
        int     text_sequence;
        char    text_segment[2048];
        char    the_rule[16000];
    EXEC SQL end declare section;

    FILE *rulef;

    rulef = fopen(the_rulefilename, "w");
    if (rulef == NULL)
        panic("Couldn't open rules-file", 'Y');

    the_rule[0] = '\0';

    EXEC SQL
        select  text_sequence, text_segment
        into    :text_sequence, :text_segment
        from    iirules
        where   table_name = :the_table
          and   rule_owner = :the_owner ;

    EXEC SQL begin;

        if (text_sequence == 1)
        {
            if (the_rule[0] != '\0')
            {
                fprintf(rulef, "%s\t%s\n", the_owner, the_rule);
                rules_saved++;
            }
            the_rule[0] = '\0';
        }

        strcat(the_rule, text_segment);

    EXEC SQL end;
    checkerrors("Error selecting from IIrules", 'Y');

    if (the_rule[0] != '\0') {
        fprintf(rulef, "%s\t%s\n", the_owner, the_rule);
        rules_saved++;
    }

    fclose(rulef);
}



void save_integrities()
{
    EXEC SQL begin declare section;
        int     integrity_number, text_sequence, prev_integrity_number;
        char    text_segment[2048];
        char    the_integrity[16000];
    EXEC SQL end declare section;

    FILE *integf;

    integf = fopen(the_integfilename, "w");
    if (integf == NULL)
        panic("Couldn't open the integrities file for output", 'Y');

    the_integrity[0] = '\0';
    prev_integrity_number = 0;

    EXEC SQL
        select  integrity_number, text_sequence, text_segment
        into    :integrity_number, :text_sequence, :text_segment
        from    iiintegrities
        where   table_name = :the_table
          and   table_owner = :the_owner
        order by integrity_number, text_sequence;

    EXEC SQL begin;

        if (integrity_number != prev_integrity_number)
        {
            if (the_integrity[0] != '\0') {
                fprintf(integf, "%s\t%s\n", the_owner, the_integrity);
                integrities_saved++;
            }

            the_integrity[0] = '\0';
        }

        strcat(the_integrity, text_segment);
        prev_integrity_number = integrity_number;

    EXEC SQL end;
    checkerrors("Error selecting from IIintegrities", 'Y');

    if (the_integrity[0] != '\0') {
        fprintf(integf, "%s\t%s\n", the_owner, the_integrity);
        integrities_saved++;
    }

    fclose(integf);
}



/*
| The save_views() routine retrieves all the views dependant upon the table,
| selects the "text" for creating the view, and writes it out the the file.
*/
void save_views()
{
    EXEC SQL begin declare section;
        int     text_sequence, i;
        char    view_name[40], view_owner[40];
        char    text_segment[2048], the_view[16000];
        char    check_option[2], view_all_to_all[2], view_ret_to_all[2];
    EXEC SQL end declare section;

    FILE * viewf;

    viewf = fopen(the_viewfilename, "w");
    if (viewf == NULL)
        panic("Couldn't open views file for output", 'Y');

    get_all_views(the_table, the_owner);

    qsort(view_arr, view_arr_elems, sizeof(VIEW_REC),
                (int (*)(const void *, const void *)) view_compare);

    for (i = 0; i < view_arr_elems; i++)
    {
        strcpy(view_name, view_arr[i].view_name);
        strcpy(view_owner, view_arr[i].view_owner);
        the_view[0] = '\0';

        EXEC SQL REPEATED
                select  check_option, text_sequence, text_segment
                into    :check_option, :text_sequence, :text_segment
                from    iiviews
                where   table_name = :view_name
                  and   table_owner = :view_owner
                order by text_sequence;

        EXEC SQL begin;

                strcat(the_view, text_segment);

        EXEC SQL end;

        checkerrors("Error selecting from IIviews", 'Y');
        sqlca.sqlcode = 0;

        /*
        | there should ALWAYS be a IIviews row-set for a view, but just
        | to be cautious, make sure it's not empty!
        */
        if (the_view[0] != '\0') {
            if (check_option[0] == 'Y') strcat(the_view, " with check option");
            fprintf(viewf, "%s\t%s\n", view_owner, the_view);
            views_saved++;


            /* good lord, you can have permissions on a VIEW as well! */
            EXEC SQL REPEATED
                select  squeeze(all_to_all), squeeze(ret_to_all)
                into    :view_all_to_all, :view_ret_to_all
                from    iitables
                where   table_name = :view_name
                  and   table_owner = :view_owner
                  and   table_type = 'V';

            checkerrors("Error selecting VIEW permission from IItables", 'Y');

            if (view_all_to_all[0] == 'Y')
                fprintf(viewf, "%s\tgrant all on %s to public\n",
                                view_owner, view_name);

            if (view_ret_to_all[0] == 'Y')
                fprintf(viewf, "%s\tgrant select on %s to public\n",
                                view_owner, view_name);

            save_permits(view_name, view_owner);
            sqlca.sqlcode = 0;
        }
    }
    checkerrors("Error fetching from IIrelation/IIdbdepends", 'Y');

    fclose(viewf);
}


int lookup_view_elem(char *vn, char *vo)
{
    int i;

    for (i = 0; i < view_arr_elems; i++)
    {
        if ((strcmp(view_arr[i].view_name, vn) == 0)    &&
            (strcmp(view_arr[i].view_owner, vo) == 0))
                return (i);
    }

    return (-1);
}



/*
| This procedure recursively finds ALL views dependant on specified
| object/owner.  Many thanks to Per Oyvind Overli, for the code
| (that I've gone and butchered)!
*/
void get_all_views(char *p_obj_name, char *p_obj_owner)
{
    EXEC SQL begin declare section;
        char    *obj_name = p_obj_name;
        char    *obj_owner = p_obj_owner;
        char    view_name[40], view_owner[40];
        int     view_reltid, i, x;
    EXEC SQL end declare section;

    /*
    | Since the select to find dependant views can be fairly expensive
    | (especially on an old VAX), do a quick check before-hand, to make
    | sure that the view hasn't already been seen ...
    */
    x = lookup_view_elem(obj_name, obj_owner);
    if (x >= 0) {
        if (view_arr[x].got_sub_views == 1)
            return;
        else
            /* don't *truly* have them until the select-loop below finishes! */
            view_arr[x].got_sub_views = 1;
    }

    EXEC SQL REPEATED
        select  squeeze(b.relid), squeeze(b.relowner), b.reltid
        into    :view_name, :view_owner, :view_reltid
        from    iirelation a, iidbdepends d, iirelation b
        where   a.relid         = :obj_name
          and   a.relowner      = :obj_owner
          and   a.reltid        = d.inid1
          and   a.reltidx       = d.inid2
          and   d.deid1         = b.reltid
          and   d.deid2         = b.reltidx
          and   b.relowner     != '$ingres';

    EXEC SQL begin;

        /*
        | we have a dependant view.  If it hasn't already been added to
        | the view_arr list, then add it.
        */
        if (lookup_view_elem(view_name, view_owner) == -1)
        {
            strcpy(view_arr[view_arr_elems].view_name, view_name);
            strcpy(view_arr[view_arr_elems].view_owner, view_owner);
            view_arr[view_arr_elems].view_reltid = view_reltid;
            view_arr[view_arr_elems].got_sub_views = 0;
            view_arr_elems++;
        }

    EXEC SQL end;
    checkerrors("Error selecting dependant views", 'Y');
    sqlca.sqlcode = 0;

    /*
    | Make sure that ALL views are checked for dependancies.  Note that
    | there's a bit of code at the start of this procedure that checks
    | to make sure a view hasn't already been checked...
    */
    for (i = 0; i < view_arr_elems; i++)
        get_all_views(view_arr[i].view_name, view_arr[i].view_owner);
}



/*
| The save_permits() routine retrieves all the rows from the IIpermits table
| and writes them out to a file.  Note that this is called for both the "base"
| table and any views based on the table.
*/
void save_permits(obj_name, obj_owner)
  EXEC SQL begin declare section;
  char *obj_name, *obj_owner;
  EXEC SQL end declare section;
{
    EXEC SQL begin declare section;
        char    text_segment[256], the_perm[16000], open_mode[2];
        int     permit_number, prev_permit_number, text_sequence;
    EXEC SQL end declare section;

    FILE *permf;


    /*
    | if the parameter "obj_name" is the same as the "base" table, we wanna
    | create the permits file.  If the_object is different from base table,
    | then we just want to append to the file...
    | [so just make sure save_permits() is always called PRIOR to save_views()
    */
    if (strcmp(obj_name, the_table) == 0)
        strcpy(open_mode, "w");
    else
        strcpy(open_mode, "a");

    strcpy(open_mode, "w");
    permf = fopen(the_permfilename, open_mode);
    if (permf == NULL)
        panic("cannot open permits output file", 'Y');

    prev_permit_number = -1;
    the_perm[0] = '\0';

    EXEC SQL REPEATED
        select  permit_number, text_sequence, text_segment
        into    :permit_number, :text_sequence, :text_segment
        from    iipermits
        where   object_name = :obj_name
          and   object_owner = :obj_owner
        order by permit_number, text_sequence;

    EXEC SQL begin;

        if ((permit_number != prev_permit_number) && (the_perm[0] != '\0'))
        {
            fprintf(permf, "%s\t%s\n", obj_owner, the_perm);
            the_perm[0] = '\0';
            permits_saved++;
        }

        strcat(the_perm, text_segment);
        prev_permit_number = permit_number;

    EXEC SQL end;
    checkerrors("Error selecting from IIpermits", 'Y');

    if (the_perm[0] != '\0') {
        fprintf(permf, "%s\t%s\n", obj_owner, the_perm);
        permits_saved++;
    }

    fclose(permf);
}



int restore_rules()
{
    if (rules_saved != 0) {
        printf("restoring %d rules\n", rules_saved);
        return restore_stuff(the_rulefilename);
    }

    return 0;
}


int restore_views()
{
    if (views_saved != 0) {
        printf("restoring %d views\n", views_saved);
        return restore_stuff(the_viewfilename);
    }

    return 0;
}


int restore_integrities()
{
    if (integrities_saved != 0) {
        printf("restoring %d integrities\n", integrities_saved);
        return restore_stuff(the_integfilename);
    }

    return 0;
}


int restore_permits()
{
    if (permits_saved != 0) {
        printf("restoring %d permits\n", permits_saved);
        return restore_stuff(the_permfilename);
    }

    return 0;
}


int do_post_sql_stuff()
{
    if (post_sql_saved != 0) {
        printf("running post_sql commands\n");
        return restore_stuff(the_psqlfilename);
    }

    return 0;
}



int restore_stuff(char *fname)
{
    EXEC SQL begin declare section;
        char    the_line[16000], *as_user, *sql_stmt;
        int     lines, errors_found;
    EXEC SQL end declare section;

    FILE *inf;

    errors_found = 0;
    lines = 0;
    inf = fopen(fname, "r");
    if (inf == NULL)
    {
        sprintf(msg, "Could not open '%s' for input", fname);
        panic(msg, 'N');  /* <- panic with 'N', cause we DON'T want to */
        return -1;                /* abort [the table's already been modified] */
    }

    /* read the file, and execute each line */
    while (fgets(the_line, sizeof(the_line), inf))
    {
        lines++;
        as_user = strtok(the_line, "\t");
        sql_stmt = strtok(NULL, "\n");

        if (strcmp(as_user, curr_user) != 0)
        {
            connect_as(as_user, 'N');
            if (ingerr != 0)
            {
                printf("skipping SQL statement:\n%s\n\n", sql_stmt);
                fprintf(logf, "skipping SQL statement:\n%s\n\n", sql_stmt);
                continue;
            }

            strcpy(curr_user, as_user);
        }


        do_sql(sql_stmt, 'N');
        if (ingerr != 0) errors_found++;
        EXEC SQL commit;
    }

    fclose(inf);

    return(errors_found);
}
